import { Route, RouteProps } from 'react-router-dom'
import { appPrivateRoutes, appPublicRoutes } from '.'
import { CreatePage } from '../pages/create'
import { HomePage } from '../pages/home'
import { LandingPage } from '../pages/landing'
import { ProfilePage } from '../pages/profile'
import { NostrLoginPage } from '../pages/settings/nostrLogin'
import { ProfileSettingsPage } from '../pages/settings/profile'
import { RelaysPage } from '../pages/settings/relays'
import { SettingsLayout } from '../pages/settings/Settings'
import { SignPage } from '../pages/sign'
import { VerifyPage } from '../pages/verify'
import { PrivateRoute } from './PrivateRoute'

/**
 * Helper type allows for extending react-router-dom's **RouteProps** with generic type
 */
type CustomRouteProps<T> = T &
  Omit<RouteProps, 'children'> & {
    children?: Array<CustomRouteProps<T>>
  }

/**
 *  This function maps over nested routes with optional condition for rendering
 *  @param {CustomRouteProps<T>[]} routes - routes list
 *  @param {RenderConditionCallbackFn} renderConditionCallbackFn - render condition callback (default true)
 */
export function recursiveRouteRenderer<T>(
  routes?: CustomRouteProps<T>[],
  renderConditionCallbackFn: (route: CustomRouteProps<T>) => boolean = () =>
    true
) {
  if (!routes) return null

  // Callback allows us to pass arbitrary conditions for each route's rendering
  // Skipping the callback will by default evaluate to true (show route)
  return routes.map((route, index) =>
    renderConditionCallbackFn(route) ? (
      <Route
        key={route.path ? `${route.path}${index}` : index}
        path={route.path}
        element={route.element}
      >
        {recursiveRouteRenderer(route.children, renderConditionCallbackFn)}
      </Route>
    ) : null
  )
}

type PublicRouteProps = CustomRouteProps<{
  hiddenWhenLoggedIn?: boolean
}>

export const publicRoutes: PublicRouteProps[] = [
  {
    path: appPublicRoutes.landingPage,
    hiddenWhenLoggedIn: true,
    element: <LandingPage />
  },
  {
    path: appPublicRoutes.profile,
    element: <ProfilePage />
  },
  {
    path: `${appPublicRoutes.verify}/:id?`,
    element: <VerifyPage />
  }
]

export const privateRoutes: CustomRouteProps<unknown>[] = [
  {
    path: appPrivateRoutes.homePage,
    element: (
      <PrivateRoute>
        <HomePage />
      </PrivateRoute>
    )
  },
  {
    path: appPrivateRoutes.create,
    element: (
      <PrivateRoute>
        <CreatePage />
      </PrivateRoute>
    )
  },
  {
    path: `${appPrivateRoutes.sign}/:id?`,
    element: (
      <PrivateRoute>
        <SignPage />
      </PrivateRoute>
    )
  },
  {
    element: (
      <PrivateRoute>
        <SettingsLayout />
      </PrivateRoute>
    ),
    children: [
      {
        path: appPrivateRoutes.profileSettings,
        element: <ProfileSettingsPage />
      },
      {
        path: appPrivateRoutes.relays,
        element: <RelaysPage />
      },
      {
        path: appPrivateRoutes.nostrLogin,
        element: <NostrLoginPage />
      }
    ]
  }
]
